/*
** Copyright (C) 2003 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "mainwindow.h"
#include "libraries.h"
#include "dialogopen.h"
#include "dialogconvert2web.h"
#include "dialogshowscript.h"
#include "constants.h"
#include "utilities.h"
#include "htmlbrowser.h"
#include "usfmtools.h"
#include "dialogfind.h"
#include "dialogreplace.h"
#include "log.h"
#include "dialogcompanion.h"
#include "dialoggotoreference.h"
#include <signal.h>
#include "bible.h"
#include "projectutils.h"
#include "dialogproject.h"
#include "dialogprojects.h"
#include "languageutils.h"
#include "dialoglanguages.h"
#include "stylesheetutils.h"
#include "dialogstylesheets.h"
#include "usfm.h"
#include "dialogimportsfm.h"
#include "worddatabase.h"
#include "dialogreplacing.h"
#include <gdk/gdkkeysyms.h>
#include <glib.h>
#include "dialogsearchspecial.h"
#include "referenceutils.h"
#include <config.h>
#include "references.h"


/*
|
|
|
|
|
Construction and destruction
|
|
|
|
|
 */


MainWindow::MainWindow ():
track (0),
//backup (0), 
  undo (0), windowscompanionlocal (true), windowscompanionremote (true)
{
  // Entry in the log.
  log_message ("Bibledit starts");

  accel_group = gtk_accel_group_new ();

  mainwindow = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (mainwindow), "Bible Editor");
  gtk_window_set_default_size (GTK_WINDOW (mainwindow), 800, 600);
  gtk_window_set_default_icon_from_file (g_build_filename (PACKAGE_DATA_DIR, "icon.xpm", NULL),
                                         NULL);

  // Initialize configuration here, only after mainwindow has been initialized.
  configuration = new Configuration (GTK_WINDOW (mainwindow));

  // Size and position of the main window.
  gtk_window_resize (GTK_WINDOW (mainwindow), configuration->window_width,
                     configuration->window_height);
  gtk_window_move (GTK_WINDOW (mainwindow), configuration->window_x_position,
                   configuration->window_y_position);

  vbox1 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox1);
  gtk_container_add (GTK_CONTAINER (mainwindow), vbox1);

  menubar1 = gtk_menu_bar_new ();
  gtk_widget_show (menubar1);
  gtk_box_pack_start (GTK_BOX (vbox1), menubar1, FALSE, FALSE, 0);

  menuitem_file = gtk_menu_item_new_with_mnemonic ("_File");
  gtk_widget_show (menuitem_file);
  gtk_container_add (GTK_CONTAINER (menubar1), menuitem_file);

  menuitem_file_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem_file), menuitem_file_menu);

  file_project = gtk_image_menu_item_new_with_mnemonic ("_Project");
  gtk_widget_show (file_project);
  gtk_container_add (GTK_CONTAINER (menuitem_file_menu), file_project);

  image463 = gtk_image_new_from_stock ("gtk-index", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image463);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (file_project), image463);

  file_project_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (file_project), file_project_menu);

  new1 = gtk_image_menu_item_new_from_stock ("gtk-new", accel_group);
  gtk_widget_show (new1);
  gtk_container_add (GTK_CONTAINER (file_project_menu), new1);

  open1 = gtk_image_menu_item_new_from_stock ("gtk-open", accel_group);
  gtk_widget_show (open1);
  gtk_container_add (GTK_CONTAINER (file_project_menu), open1);

  close1 = gtk_image_menu_item_new_from_stock ("gtk-close", accel_group);
  gtk_widget_show (close1);
  gtk_container_add (GTK_CONTAINER (file_project_menu), close1);

  delete1 = gtk_image_menu_item_new_from_stock ("gtk-delete", accel_group);
  gtk_widget_show (delete1);
  gtk_container_add (GTK_CONTAINER (file_project_menu), delete1);

  properties1 = gtk_image_menu_item_new_from_stock ("gtk-properties", accel_group);
  gtk_widget_show (properties1);
  gtk_container_add (GTK_CONTAINER (file_project_menu), properties1);

  import1 = gtk_image_menu_item_new_with_mnemonic ("_Import");
  gtk_widget_show (import1);
  gtk_container_add (GTK_CONTAINER (file_project_menu), import1);

  image464 = gtk_image_new_from_stock ("gtk-convert", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image464);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (import1), image464);

  file_references = gtk_image_menu_item_new_with_mnemonic ("_References");
  gtk_widget_show (file_references);
  gtk_container_add (GTK_CONTAINER (menuitem_file_menu), file_references);

  image465 = gtk_image_new_from_stock ("gtk-find", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image465);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (file_references), image465);

  file_references_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (file_references), file_references_menu);

  open_references1 = gtk_image_menu_item_new_with_mnemonic ("_Open");
  gtk_widget_show (open_references1);
  gtk_container_add (GTK_CONTAINER (file_references_menu), open_references1);

  image466 = gtk_image_new_from_stock ("gtk-open", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image466);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (open_references1), image466);

  references_save_as = gtk_image_menu_item_new_from_stock ("gtk-save-as", accel_group);
  gtk_widget_show (references_save_as);
  gtk_container_add (GTK_CONTAINER (file_references_menu), references_save_as);

  close_references = gtk_image_menu_item_new_with_mnemonic ("_Close");
  gtk_widget_show (close_references);
  gtk_container_add (GTK_CONTAINER (file_references_menu), close_references);

  image468 = gtk_image_new_from_stock ("gtk-close", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image468);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (close_references), image468);

  delete_references = gtk_image_menu_item_new_with_mnemonic ("_Delete");
  gtk_widget_show (delete_references);
  gtk_container_add (GTK_CONTAINER (file_references_menu), delete_references);

  image469 = gtk_image_new_from_stock ("gtk-delete", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image469);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (delete_references), image469);

  quit1 = gtk_image_menu_item_new_from_stock ("gtk-quit", accel_group);
  gtk_widget_show (quit1);
  gtk_container_add (GTK_CONTAINER (menuitem_file_menu), quit1);

//    menufile.items ().push_back (MenuElem ("_Test", sigc::mem_fun (*this, &MainWindow::on_test)));

  menuitem_edit = gtk_menu_item_new_with_mnemonic ("_Edit");
  gtk_widget_show (menuitem_edit);
  gtk_container_add (GTK_CONTAINER (menubar1), menuitem_edit);

  menuitem_edit_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem_edit), menuitem_edit_menu);

  cut1 = gtk_image_menu_item_new_from_stock ("gtk-cut", accel_group);
  gtk_widget_show (cut1);
  gtk_container_add (GTK_CONTAINER (menuitem_edit_menu), cut1);

  copy1 = gtk_image_menu_item_new_from_stock ("gtk-copy", accel_group);
  gtk_widget_show (copy1);
  gtk_container_add (GTK_CONTAINER (menuitem_edit_menu), copy1);

  paste1 = gtk_image_menu_item_new_from_stock ("gtk-paste", accel_group);
  gtk_widget_show (paste1);
  gtk_container_add (GTK_CONTAINER (menuitem_edit_menu), paste1);

  separator2 = gtk_separator_menu_item_new ();
  gtk_widget_show (separator2);
  gtk_container_add (GTK_CONTAINER (menuitem_edit_menu), separator2);
  gtk_widget_set_sensitive (separator2, FALSE);

  undo1 = gtk_image_menu_item_new_with_mnemonic ("_Undo");
  gtk_widget_show (undo1);
  gtk_container_add (GTK_CONTAINER (menuitem_edit_menu), undo1);
  gtk_widget_add_accelerator (undo1, "activate", accel_group, GDK_Z, GDK_CONTROL_MASK,
                              GTK_ACCEL_VISIBLE);

  image295 = gtk_image_new_from_stock ("gtk-undo", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image295);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (undo1), image295);

  redo1 = gtk_image_menu_item_new_with_mnemonic ("_Redo");
  gtk_widget_show (redo1);
  gtk_container_add (GTK_CONTAINER (menuitem_edit_menu), redo1);
  gtk_widget_add_accelerator (redo1, "activate", accel_group, GDK_Z,
                              GdkModifierType (GDK_CONTROL_MASK | GDK_SHIFT_MASK),
                              GTK_ACCEL_VISIBLE);

  image296 = gtk_image_new_from_stock ("gtk-redo", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image296);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (redo1), image296);

  separator4 = gtk_separator_menu_item_new ();
  gtk_widget_show (separator4);
  gtk_container_add (GTK_CONTAINER (menuitem_edit_menu), separator4);
  gtk_widget_set_sensitive (separator4, FALSE);

  find1 = gtk_image_menu_item_new_from_stock ("gtk-find", accel_group);
  gtk_widget_show (find1);
  gtk_container_add (GTK_CONTAINER (menuitem_edit_menu), find1);

  find_and_replace1 = gtk_image_menu_item_new_from_stock ("gtk-find-and-replace", accel_group);
  gtk_widget_show (find_and_replace1);
  gtk_container_add (GTK_CONTAINER (menuitem_edit_menu), find_and_replace1);

  findspecial1 = gtk_image_menu_item_new_with_mnemonic ("Find _Special");
  gtk_widget_show (findspecial1);
  gtk_container_add (GTK_CONTAINER (menuitem_edit_menu), findspecial1);

  image313 = gtk_image_new_from_stock ("gtk-find", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image313);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (findspecial1), image313);

  menuitem_view = gtk_menu_item_new_with_mnemonic ("_View");
  gtk_widget_show (menuitem_view);
  gtk_container_add (GTK_CONTAINER (menubar1), menuitem_view);

  menuitem_view_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem_view), menuitem_view_menu);

  font1 = gtk_image_menu_item_new_from_stock ("gtk-select-font", accel_group);
  gtk_widget_show (font1);
  gtk_container_add (GTK_CONTAINER (menuitem_view_menu), font1);

  menuitem_goto = gtk_menu_item_new_with_mnemonic ("_Goto");
  gtk_widget_show (menuitem_goto);
  gtk_container_add (GTK_CONTAINER (menubar1), menuitem_goto);

  menuitem_goto_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem_goto), menuitem_goto_menu);

  next_verse1 = gtk_image_menu_item_new_with_mnemonic ("Next verse");
  gtk_widget_show (next_verse1);
  gtk_container_add (GTK_CONTAINER (menuitem_goto_menu), next_verse1);
  gtk_widget_add_accelerator (next_verse1, "activate", accel_group, GDK_F7, GdkModifierType (0),
                              GTK_ACCEL_VISIBLE);
  // The GdkModifierType was given as 0 by glade's code, but that did not compile.
  // Here we take a 0 and cast it to a GdkModifierType. That works well.

  image95 = gtk_image_new_from_stock ("gtk-go-forward", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image95);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (next_verse1), image95);

  previous_verse1 = gtk_image_menu_item_new_with_mnemonic ("Previous verse");
  gtk_widget_show (previous_verse1);
  gtk_container_add (GTK_CONTAINER (menuitem_goto_menu), previous_verse1);
  gtk_widget_add_accelerator (previous_verse1, "activate", accel_group, GDK_F7, GDK_SHIFT_MASK,
                              GTK_ACCEL_VISIBLE);

  image96 = gtk_image_new_from_stock ("gtk-go-back", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image96);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (previous_verse1), image96);

  next_chapter1 = gtk_image_menu_item_new_with_mnemonic ("Next chapter");
  gtk_widget_show (next_chapter1);
  gtk_container_add (GTK_CONTAINER (menuitem_goto_menu), next_chapter1);
  gtk_widget_add_accelerator (next_chapter1, "activate", accel_group, GDK_F8, GdkModifierType (0),
                              GTK_ACCEL_VISIBLE);

  image97 = gtk_image_new_from_stock ("gtk-go-forward", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image97);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (next_chapter1), image97);

  previous_chapter1 = gtk_image_menu_item_new_with_mnemonic ("Previous chapter");
  gtk_widget_show (previous_chapter1);
  gtk_container_add (GTK_CONTAINER (menuitem_goto_menu), previous_chapter1);
  gtk_widget_add_accelerator (previous_chapter1, "activate", accel_group, GDK_F8, GDK_SHIFT_MASK,
                              GTK_ACCEL_VISIBLE);

  image98 = gtk_image_new_from_stock ("gtk-go-back", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image98);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (previous_chapter1), image98);

  next_book1 = gtk_image_menu_item_new_with_mnemonic ("Next book");
  gtk_widget_show (next_book1);
  gtk_container_add (GTK_CONTAINER (menuitem_goto_menu), next_book1);
  gtk_widget_add_accelerator (next_book1, "activate", accel_group, GDK_F9, GdkModifierType (0),
                              GTK_ACCEL_VISIBLE);

  image99 = gtk_image_new_from_stock ("gtk-go-forward", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image99);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (next_book1), image99);

  previous_book1 = gtk_image_menu_item_new_with_mnemonic ("Previous book");
  gtk_widget_show (previous_book1);
  gtk_container_add (GTK_CONTAINER (menuitem_goto_menu), previous_book1);
  gtk_widget_add_accelerator (previous_book1, "activate", accel_group, GDK_F9, GDK_SHIFT_MASK,
                              GTK_ACCEL_VISIBLE);

  image100 = gtk_image_new_from_stock ("gtk-go-back", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image100);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (previous_book1), image100);

  next_reference1 = gtk_image_menu_item_new_with_mnemonic ("Next reference");
  gtk_widget_show (next_reference1);
  gtk_container_add (GTK_CONTAINER (menuitem_goto_menu), next_reference1);
  gtk_widget_add_accelerator (next_reference1, "activate", accel_group, GDK_F6, GdkModifierType (0),
                              GTK_ACCEL_VISIBLE);

  image608 = gtk_image_new_from_stock ("gtk-go-down", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image608);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (next_reference1), image608);

  previous_reference1 = gtk_image_menu_item_new_with_mnemonic ("Previous reference");
  gtk_widget_show (previous_reference1);
  gtk_container_add (GTK_CONTAINER (menuitem_goto_menu), previous_reference1);
  gtk_widget_add_accelerator (previous_reference1, "activate", accel_group, GDK_F6, GDK_SHIFT_MASK,
                              GTK_ACCEL_VISIBLE);

  image609 = gtk_image_new_from_stock ("gtk-go-up", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image609);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (previous_reference1), image609);

  reference1 = gtk_image_menu_item_new_with_mnemonic ("Reference");
  gtk_widget_show (reference1);
  gtk_container_add (GTK_CONTAINER (menuitem_goto_menu), reference1);
  gtk_widget_add_accelerator (reference1, "activate", accel_group, GDK_g, GDK_CONTROL_MASK,
                              GTK_ACCEL_VISIBLE);

  image101 = gtk_image_new_from_stock ("gtk-jump-to", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image101);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (reference1), image101);

  menuitem_tools = gtk_menu_item_new_with_mnemonic ("_Tools");
  gtk_widget_show (menuitem_tools);
  gtk_container_add (GTK_CONTAINER (menubar1), menuitem_tools);

  menuitem_tools_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem_tools), menuitem_tools_menu);

  convert_project_to_web_pages1 =
    gtk_image_menu_item_new_with_mnemonic ("_Convert project to web pages");
  gtk_widget_show (convert_project_to_web_pages1);
  gtk_container_add (GTK_CONTAINER (menuitem_tools_menu), convert_project_to_web_pages1);

  image102 = gtk_image_new_from_stock ("gtk-convert", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image102);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (convert_project_to_web_pages1), image102);

  menuitem_preferences = gtk_menu_item_new_with_mnemonic ("P_references");
  gtk_widget_show (menuitem_preferences);
  gtk_container_add (GTK_CONTAINER (menubar1), menuitem_preferences);

  menuitem_preferences_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem_preferences), menuitem_preferences_menu);

  bibledit_windows_companion1 =
    gtk_image_menu_item_new_with_mnemonic ("_Bibledit Windows Companion");
  gtk_widget_show (bibledit_windows_companion1);
  gtk_container_add (GTK_CONTAINER (menuitem_preferences_menu), bibledit_windows_companion1);

  image103 = gtk_image_new_from_stock ("gtk-network", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image103);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (bibledit_windows_companion1), image103);

  languages1 = gtk_image_menu_item_new_with_mnemonic ("_Languages");
  gtk_widget_show (languages1);
  gtk_container_add (GTK_CONTAINER (menuitem_preferences_menu), languages1);

  image104 = gtk_image_new_from_stock ("gtk-spell-check", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image104);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (languages1), image104);

  stylesheets1 = gtk_image_menu_item_new_with_mnemonic ("_Stylesheets");
  gtk_widget_show (stylesheets1);
  gtk_container_add (GTK_CONTAINER (menuitem_preferences_menu), stylesheets1);

  image105 = gtk_image_new_from_stock ("gtk-select-font", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image105);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (stylesheets1), image105);

  menuitem_diagnostics = gtk_menu_item_new_with_mnemonic ("_Diagnostics");
  gtk_widget_show (menuitem_diagnostics);
  gtk_container_add (GTK_CONTAINER (menubar1), menuitem_diagnostics);

  menuitem_diagnostics_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem_diagnostics), menuitem_diagnostics_menu);

  view_system_log1 = gtk_image_menu_item_new_with_mnemonic ("_View system log");
  gtk_widget_show (view_system_log1);
  gtk_container_add (GTK_CONTAINER (menuitem_diagnostics_menu), view_system_log1);

  image106 = gtk_image_new_from_stock ("gtk-dialog-info", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image106);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (view_system_log1), image106);

  menuitem_help = gtk_menu_item_new_with_mnemonic ("_Help");
  gtk_widget_show (menuitem_help);
  gtk_container_add (GTK_CONTAINER (menubar1), menuitem_help);

  menuitem_help_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem_help), menuitem_help_menu);

  help1 = gtk_image_menu_item_new_from_stock ("gtk-help", accel_group);
  gtk_widget_show (help1);
  gtk_container_add (GTK_CONTAINER (menuitem_help_menu), help1);

  about1 = gtk_image_menu_item_new_with_mnemonic ("_About");
  gtk_widget_show (about1);
  gtk_container_add (GTK_CONTAINER (menuitem_help_menu), about1);

  image107 = gtk_image_new_from_stock ("gtk-dialog-info", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image107);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (about1), image107);

  toolbar1 = gtk_toolbar_new ();
  gtk_widget_show (toolbar1);
  gtk_box_pack_start (GTK_BOX (vbox1), toolbar1, FALSE, FALSE, 0);
  gtk_toolbar_set_style (GTK_TOOLBAR (toolbar1), GTK_TOOLBAR_BOTH);

  button_back = gtk_button_new ();
  gtk_widget_show (button_back);
  gtk_toolbar_append_element (GTK_TOOLBAR (toolbar1), GTK_TOOLBAR_CHILD_WIDGET, button_back, "",
                              NULL, NULL, NULL, NULL, NULL);

  image1 = gtk_image_new_from_stock ("gtk-go-back", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image1);
  gtk_container_add (GTK_CONTAINER (button_back), image1);

  button_forward = gtk_button_new ();
  gtk_widget_show (button_forward);
  gtk_toolbar_append_element (GTK_TOOLBAR (toolbar1), GTK_TOOLBAR_CHILD_WIDGET, button_forward, "",
                              NULL, NULL, NULL, NULL, NULL);

  image2 = gtk_image_new_from_stock ("gtk-go-forward", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image2);
  gtk_container_add (GTK_CONTAINER (button_forward), image2);

  combo_book = gtk_combo_new ();
  gtk_widget_show (combo_book);
  gtk_toolbar_append_widget (GTK_TOOLBAR (toolbar1), combo_book, NULL, NULL);

  combo_entry_book = GTK_COMBO (combo_book)->entry;
  gtk_widget_show (combo_entry_book);
  gtk_editable_set_editable (GTK_EDITABLE (combo_entry_book), FALSE);

  combo_chapter = gtk_combo_new ();
  gtk_widget_show (combo_chapter);
  gtk_toolbar_append_widget (GTK_TOOLBAR (toolbar1), combo_chapter, NULL, NULL);

  combo_entry_chapter = GTK_COMBO (combo_chapter)->entry;
  gtk_widget_show (combo_entry_chapter);
  gtk_widget_set_size_request (combo_entry_chapter, 55, -1);
  gtk_editable_set_editable (GTK_EDITABLE (combo_entry_chapter), FALSE);

  combo_verse = gtk_combo_new ();
  gtk_widget_show (combo_verse);
  gtk_toolbar_append_widget (GTK_TOOLBAR (toolbar1), combo_verse, NULL, NULL);

  combo_entry_verse = GTK_COMBO (combo_verse)->entry;
  gtk_widget_show (combo_entry_verse);
  gtk_widget_set_size_request (combo_entry_verse, 55, -1);
  gtk_editable_set_editable (GTK_EDITABLE (combo_entry_verse), FALSE);

  hbox1 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox1);
  gtk_box_pack_start (GTK_BOX (vbox1), hbox1, TRUE, TRUE, 0);

  hpaned1 = gtk_hpaned_new ();
  gtk_widget_show (hpaned1);
  gtk_box_pack_start (GTK_BOX (hbox1), hpaned1, TRUE, TRUE, 0);
  // Position of horizontal pane.
  gtk_paned_set_position (GTK_PANED (hpaned1), configuration->hpane_position);

  vpaned1 = gtk_vpaned_new ();
  gtk_widget_show (vpaned1);
  gtk_paned_pack1 (GTK_PANED (hpaned1), vpaned1, FALSE, TRUE);
  // Position of vertical pane.
  gtk_paned_set_position (GTK_PANED (vpaned1), configuration->vpane_position);

  scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow1);
  gtk_paned_pack1 (GTK_PANED (vpaned1), scrolledwindow1, FALSE, TRUE);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow1), GTK_POLICY_AUTOMATIC,
                                  GTK_POLICY_AUTOMATIC);

  textview_edit = gtk_text_view_new ();
  gtk_widget_show (textview_edit);
  gtk_container_add (GTK_CONTAINER (scrolledwindow1), textview_edit);
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (textview_edit), GTK_WRAP_WORD);

  scrolledwindow3 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow3);
  gtk_paned_pack2 (GTK_PANED (vpaned1), scrolledwindow3, TRUE, TRUE);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow3), GTK_POLICY_AUTOMATIC,
                                  GTK_POLICY_AUTOMATIC);

  textview_notes = gtk_text_view_new ();
  gtk_widget_show (textview_notes);
  gtk_container_add (GTK_CONTAINER (scrolledwindow3), textview_notes);
  gtk_text_view_set_editable (GTK_TEXT_VIEW (textview_notes), FALSE);

  scrolledwindow2 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow2);
  gtk_paned_pack2 (GTK_PANED (hpaned1), scrolledwindow2, TRUE, TRUE);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow2), GTK_POLICY_AUTOMATIC,
                                  GTK_POLICY_AUTOMATIC);

  // Manually added and changed.
  liststore = gtk_list_store_new (1, G_TYPE_STRING);
  // Text cell renderer.
  GtkCellRenderer *renderer;
  renderer = gtk_cell_renderer_text_new ();
  treeview1 = gtk_tree_view_new_with_model (GTK_TREE_MODEL (liststore));
  gtk_widget_show (treeview1);
  gtk_container_add (GTK_CONTAINER (scrolledwindow2), treeview1);
  gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (treeview1), TRUE);
  treecolumn = gtk_tree_view_column_new_with_attributes ("", renderer, "text", 0, NULL);
  gtk_tree_view_append_column (GTK_TREE_VIEW (treeview1), treecolumn);
  treeselect = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview1));
  gtk_tree_selection_set_mode (treeselect, GTK_SELECTION_MULTIPLE);

  statusbar1 = gtk_statusbar_new ();
  gtk_widget_show (statusbar1);
  gtk_box_pack_start (GTK_BOX (vbox1), statusbar1, FALSE, FALSE, 0);

  // Appearance of text in editor.
  editor_set_font ();

  g_signal_connect ((gpointer) mainwindow, "destroy", G_CALLBACK (gtk_main_quit), gpointer(this));
  g_signal_connect ((gpointer) mainwindow, "delete_event", G_CALLBACK (on_mainwindow_delete_event), gpointer(this));
  g_signal_connect ((gpointer) new1, "activate", G_CALLBACK (on_new1_activate), gpointer(this));
  g_signal_connect ((gpointer) open1, "activate", G_CALLBACK (on_open1_activate), gpointer(this));
  g_signal_connect ((gpointer) close1, "activate", G_CALLBACK (on_close1_activate), gpointer(this));
  g_signal_connect ((gpointer) delete1, "activate", G_CALLBACK (on_delete1_activate), gpointer(this));
  g_signal_connect ((gpointer) properties1, "activate", G_CALLBACK (on_properties1_activate), gpointer(this));
  g_signal_connect ((gpointer) import1, "activate", G_CALLBACK (on_import1_activate), gpointer(this));
  g_signal_connect ((gpointer) open_references1, "activate", G_CALLBACK (on_open_references1_activate), gpointer(this));
  g_signal_connect ((gpointer) references_save_as, "activate", G_CALLBACK (on_references_save_as_activate), gpointer(this));
  g_signal_connect ((gpointer) close_references, "activate", G_CALLBACK (on_close_references_activate), gpointer (this));
  g_signal_connect ((gpointer) delete_references, "activate", G_CALLBACK (on_delete_references_activate), gpointer (this));
  g_signal_connect ((gpointer) quit1, "activate", G_CALLBACK (on_quit1_activate), gpointer(this));
  g_signal_connect ((gpointer) menuitem_edit, "activate", G_CALLBACK (on_edit1_activate), gpointer(this));
  g_signal_connect ((gpointer) cut1, "activate", G_CALLBACK (on_cut1_activate), gpointer(this));
  g_signal_connect ((gpointer) copy1, "activate", G_CALLBACK (on_copy1_activate), gpointer(this));
  g_signal_connect ((gpointer) paste1, "activate", G_CALLBACK (on_paste1_activate), gpointer(this));
  g_signal_connect ((gpointer) undo1, "activate", G_CALLBACK (on_undo1_activate), gpointer(this));
  g_signal_connect ((gpointer) redo1, "activate", G_CALLBACK (on_redo1_activate), gpointer(this));
  g_signal_connect ((gpointer) find1, "activate", G_CALLBACK (on_find1_activate), gpointer(this));
  g_signal_connect ((gpointer) find_and_replace1, "activate", G_CALLBACK (on_find_and_replace1_activate), gpointer (this));
  g_signal_connect ((gpointer) findspecial1, "activate", G_CALLBACK (on_findspecial1_activate), gpointer(this));
  g_signal_connect ((gpointer) font1, "activate", G_CALLBACK (on_font1_activate), gpointer(this));
  g_signal_connect ((gpointer) next_verse1, "activate", G_CALLBACK (on_next_verse1_activate), gpointer (this));
  g_signal_connect ((gpointer) previous_verse1, "activate", G_CALLBACK (on_previous_verse1_activate), gpointer (this));
  g_signal_connect ((gpointer) next_chapter1, "activate", G_CALLBACK (on_next_chapter1_activate), gpointer (this));
  g_signal_connect ((gpointer) previous_chapter1, "activate", G_CALLBACK (on_previous_chapter1_activate), gpointer (this));
  g_signal_connect ((gpointer) next_book1, "activate", G_CALLBACK (on_next_book1_activate), gpointer (this));
  g_signal_connect ((gpointer) previous_book1, "activate", G_CALLBACK (on_previous_book1_activate), gpointer (this));
  g_signal_connect ((gpointer) next_reference1, "activate", G_CALLBACK (on_next_reference1_activate), gpointer(this));
  g_signal_connect ((gpointer) previous_reference1, "activate", G_CALLBACK (on_previous_reference1_activate), gpointer(this));
  g_signal_connect ((gpointer) reference1, "activate", G_CALLBACK (on_reference1_activate), gpointer(this));
  g_signal_connect ((gpointer) convert_project_to_web_pages1, "activate", G_CALLBACK (on_convert_project_to_web_pages1_activate), gpointer(this));
  g_signal_connect ((gpointer) bibledit_windows_companion1, "activate", G_CALLBACK (on_bibledit_windows_companion1_activate), gpointer(this));
  g_signal_connect ((gpointer) languages1, "activate", G_CALLBACK (on_languages1_activate), gpointer (this));
  g_signal_connect ((gpointer) stylesheets1, "activate", G_CALLBACK (on_stylesheets1_activate), gpointer(this));
  g_signal_connect ((gpointer) view_system_log1, "activate", G_CALLBACK (on_view_system_log1_activate), gpointer(this));
  g_signal_connect ((gpointer) help1, "activate", G_CALLBACK (on_help1_activate), gpointer(this));
  g_signal_connect ((gpointer) about1, "activate", G_CALLBACK (on_about1_activate), gpointer(this));
  g_signal_connect ((gpointer) button_back, "clicked", G_CALLBACK (on_button_back_clicked), gpointer(this));
  g_signal_connect ((gpointer) button_forward, "clicked", G_CALLBACK (on_button_forward_clicked), gpointer(this));
  g_signal_connect ((gpointer) combo_entry_book, "changed", G_CALLBACK (on_combo_entry_book_changed), gpointer(this));
  g_signal_connect ((gpointer) combo_entry_chapter, "changed", G_CALLBACK (on_combo_entry_chapter_changed), gpointer(this));
  g_signal_connect ((gpointer) combo_entry_verse, "changed", G_CALLBACK (on_combo_entry_verse_changed), gpointer(this));
  g_signal_connect ((gpointer) textview_edit, "move_cursor", G_CALLBACK (on_textview_edit_move_cursor), gpointer(this));
  g_signal_connect ((gpointer) textview_edit, "grab_focus", G_CALLBACK (on_textview_edit_grab_focus), gpointer(this));
  g_signal_connect ((gpointer) treeview1, "key_press_event", G_CALLBACK (on_treeview1_key_press_event), gpointer(this));
  g_signal_connect ((gpointer) treeview1, "button_press_event", G_CALLBACK (on_treeview1_button_press_event), gpointer(this));

  gtk_window_add_accel_group (GTK_WINDOW (mainwindow), accel_group);

  // Store pointer to the two text buffers to simplify later use.
  // The reference count on the buffer is not incremented; 
  // the caller of this function won't own a new reference.
  // Therefore the buffers do not need to be freed.
  textbuffer_edit = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview_edit));
  textbuffer_notes = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview_notes));

  // Signal handlers for the undo/redo function.
  g_signal_connect ((gpointer) textbuffer_edit, "insert_text", G_CALLBACK (on_textbuffer_edit_insert_text), gpointer(this));
  g_signal_connect ((gpointer) textbuffer_edit, "delete_range", G_CALLBACK (on_textbuffer_edit_delete_text), gpointer(this));

  // Tag for highlighting a new verse.
  // The colour names can be found in rgb.txt of the X system.
  // On Fedore Core 3 this is in /usr/X11R6/lib/X11/rgb.txt
  editor_line_tag = gtk_text_buffer_create_tag (textbuffer_edit, "linetag", "background", "light yellow", NULL);

  // Tag for highlighting search words
  editor_reference_tag = gtk_text_buffer_create_tag (textbuffer_edit, "referencetag", "background", "khaki", NULL);

  // Statusbar context id
  statusbar_context_id = gtk_statusbar_get_context_id (GTK_STATUSBAR (statusbar1), "Statusbar");

  // Initalize some variables.
  go_to_reference_running = false;
  scripture = NULL;
  navigator_book = NULL;
  book = NULL;

  // Project initialization.
  initialize_project ();

  // Get the system for tracking the Bible references up and running.
  g_timeout_add (1000, GSourceFunc (mainwindow_on_tracker_timeout), gpointer(this));
  set_sensitivity_of_navigator_buttons ();

  // System for navigation.
  g_timeout_add (100, GSourceFunc (mainwindow_on_navigation_timeout), gpointer(this));

  // Automatic saving of the file, periodiodically.
  g_timeout_add (60000, GSourceFunc (mainwindow_on_save_timeout), gpointer(this));

  // Communication with Windows Companion
  if (configuration->companion_local_enabled)
    windowscompanionlocal.Connect ("localhost", 51515);
  if (configuration->companion_remote_enabled)
    windowscompanionremote.Connect (configuration->companion_remote_name, 51515);
  g_timeout_add (100, GSourceFunc (mainwindow_on_windows_companion_timeout), gpointer(this));

  // Block the signal of a pipe error that otherwise would kill bibledit.
  signal (SIGPIPE, SIG_IGN);

  // Load previously saved references, if any.
  References references (liststore, treeview1, treecolumn);
  references.load ("");
  references.get_references (references_searchwords, references_casesensitives);
  references.fill_store ();

  // Finally: show main window.
  gtk_widget_show (mainwindow);
}


MainWindow::~MainWindow ()
{
  // Save the size and position of the program window.
  // The get_position function is broken, See the gtk documentation for reasons.
  // Anyway, for just now we'll leave the variables in configuration, and we hope
  // that it will be fixed later. The problem now is that the function will
  // return the value that was set with move(x,y). If the user moves the window
  // that is not reflected here, but it will keep returning the values set
  // with move(x,y).
  // If you want to change the default values when Bibledit starts, edit
  // the configuration file
  gtk_window_get_size (GTK_WINDOW (mainwindow), &configuration->window_width, &configuration->window_height);
  gtk_window_get_position (GTK_WINDOW (mainwindow), &configuration->window_x_position, &configuration->window_y_position);
  configuration->hpane_position = gtk_paned_get_position (GTK_PANED (hpaned1));
  configuration->vpane_position = gtk_paned_get_position (GTK_PANED (vpaned1));
  // Save text in editor
  editor_save ();
  // Remove system log.
  unlink (LOG_FILE);
  if (scripture)
    delete scripture;
  if (navigator_book)
    delete navigator_book;
  if (book)
    delete book;
  // Optimize word database
  if (!configuration->project.empty ())
    {
      string command = "bibledit.wdb --finalize ";
      command.append (configuration->project);
      command.append (" &");
      system (command.c_str());
    }
  if (configuration)
    delete configuration;
  // Destroying the window is done by gtk itself.
}


int
MainWindow::run ()
{
  return gtk_dialog_run (GTK_DIALOG (mainwindow));
}


/*
|
|
|
|
|
Initialization
|
|
|
|
|
 */


void
MainWindow::initialize_project ()
{

  // If we've no project, disables the menus and knock off.
  if (configuration->project.empty ())
    {
      enable_or_disable_widgets (false);
      return;
    }
  // Get the Scripture.
  scripture = new Scripture (configuration->project);

  // Update the statusbar.
  int context_id = gtk_statusbar_get_context_id (GTK_STATUSBAR (statusbar1), "");
  gtk_statusbar_push (GTK_STATUSBAR (statusbar1), context_id, configuration->project.c_str ());

  // Enable the menus.
  enable_or_disable_widgets (true);

  // Variable needed otherwise the navigation rumbles too much.
  driving_book = true;

  // Put the names of the books in the combo.
  // Put them also in a vector, because the books cannot be extracted from the widget itself.
  GList *combo_book_items = NULL;
  combo_book_popdown_strings.clear ();
  string gotobook;
  for (unsigned int i = 0; i < scripture->books.size (); i++)
    {
      combo_book_items = g_list_append (combo_book_items, (gpointer) scripture->books[i].c_str ());
      if (i == 0)
        gotobook = scripture->books[0];
      combo_book_popdown_strings.push_back (scripture->books[i]);
    }
  gtk_combo_set_popdown_strings (GTK_COMBO (combo_book), combo_book_items);
  g_list_free (combo_book_items);


  int bookindex = scripture->get_index_of_book (configuration->book_opened);
  if (bookindex >= 0)
    {
      // Set the book to the one that was opened last.
      gotobook = scripture->books[bookindex];
    }
  else
    {
      // If the book does not exist, reset chapter and verse.
      configuration->chapter_opened = "1";
      configuration->verse_opened = "1";
    }

  // Combo has been filled: clear the anti-rumble variable.
  driving_book = false;

  // Clear some variables so that the new book will be loaded.
  editor_current_book.clear ();
  editor_current_chapter.clear ();
  editor_current_verse.clear ();
  navigator_current_book.clear ();
  navigator_current_chapter.clear ();
  navigator_current_verse.clear ();

  // Go to the new reference.
  if (!gotobook.empty ())
    go_to_reference (gotobook, configuration->chapter_opened, configuration->verse_opened, true,
                     false);

  // Clear the references tracker.
  track.clear ();

  // Initialize word database.
  {
    string command = "bibledit.wdb --initialize '";
    command.append (configuration->project);
    command.append ("' &");
    system (command.c_str ());
  }
}


void
MainWindow::enable_or_disable_widgets (bool enable)
{
  // Depending on whether a project has been opened, make some widgets (in)sensitive.
  gtk_widget_set_sensitive (close1, enable);
  gtk_widget_set_sensitive (properties1, enable);
  gtk_widget_set_sensitive (import1, enable);
  gtk_widget_set_sensitive (menuitem_edit, enable);
  gtk_widget_set_sensitive (menuitem_view, enable);
  gtk_widget_set_sensitive (menuitem_goto, enable);
  gtk_widget_set_sensitive (menuitem_tools, enable);
  gtk_widget_set_sensitive (combo_book, enable);
  gtk_widget_set_sensitive (combo_chapter, enable);
  gtk_widget_set_sensitive (combo_verse, enable);
  if (!enable)
    {
      gtk_entry_set_text (GTK_ENTRY (combo_entry_book), "");
      gtk_entry_set_text (GTK_ENTRY (combo_entry_chapter), "");
      gtk_entry_set_text (GTK_ENTRY (combo_entry_verse), "");
    }
  gtk_widget_set_sensitive (textview_edit, enable);
}


/*
|
|
|
|
|
Menu callbacks
|
|
|
|
|
 */


void MainWindow::on_open1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->open ();
}


void MainWindow::open ()
{
  OpenDialog opendialog (configuration->project);
  if (opendialog.run () == GTK_RESPONSE_OK)
    {
      close ();
      configuration->project = opendialog.project;
      initialize_project ();
    }
}


void MainWindow::on_close1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->close ();
}


void
MainWindow::close ()
{
  editor_save ();
  if (book)
    {
      delete book;
      book = NULL;
    }
  enable_or_disable_widgets (false);
  gtk_text_buffer_set_text (textbuffer_edit, "", -1);
  gtk_text_buffer_set_modified (textbuffer_edit, false);
  editor_current_book.clear ();
  editor_current_chapter.clear ();
  editor_current_verse.clear ();
  navigator_current_book.clear ();
  navigator_current_chapter.clear ();
  navigator_current_verse.clear ();
  gtk_statusbar_pop (GTK_STATUSBAR (statusbar1), statusbar_context_id);
  configuration->project.clear ();
  if (scripture)
    {
      delete scripture;
      scripture = NULL;
    }
  if (navigator_book)
    {
      delete navigator_book;
      navigator_book = NULL;
    }
}


void MainWindow::on_new1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->newproject ();
}


void
MainWindow::newproject ()
{
  projectutils_new ();
  ProjectDialog projectdialog (NEW_PROJECT);
  if (projectdialog.run () == GTK_RESPONSE_OK)
    {
      close ();
      configuration->project = projectdialog.newprojectname;
      initialize_project ();
    }
}


void MainWindow::on_properties1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->editproject ();
}


void
MainWindow::editproject ()
{
  ProjectDialog projectdialog (configuration->project);
  if (projectdialog.run () == GTK_RESPONSE_OK)
    {
      close ();
      configuration->project = projectdialog.newprojectname;
      initialize_project ();
    }
}


void MainWindow::on_delete1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->deleteproject ();
}


void
MainWindow::deleteproject ()
{
  ProjectsDialog dialog (configuration->project);
  dialog.run ();
}


void MainWindow::on_quit1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  gtk_main_quit ();
}


void MainWindow::on_convert_project_to_web_pages1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->convert2web ();
}


void
MainWindow::convert2web ()
{
  ConvertToWebDialog converttoweb (configuration);
  if (converttoweb.run () == GTK_RESPONSE_OK)
    {
      configuration->converttowebdirectory = converttoweb.converttowebdirectory;
    }
}


void MainWindow::on_view_system_log1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->viewlog ();
}


void
MainWindow::viewlog ()
{
  ShowScriptDialog showscript (LOG_FILE);
  showscript.run ();
}


void MainWindow::on_help1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  htmlbrowser (g_build_filename (PACKAGE_DATA_DIR, "index.html", NULL));
}


void MainWindow::on_about1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->showabout ();
}


void
MainWindow::showabout ()
{
  string message = "bibledit ";
  message.append (PACKAGE_VERSION);
  message.append ("\n\n");
  message.append (PACKAGE_BUGREPORT);
  message.append ("\n");
  GtkWidget *dialog = gtk_message_dialog_new (GTK_WINDOW (mainwindow),
                                              GTK_DIALOG_DESTROY_WITH_PARENT,
                                              GTK_MESSAGE_INFO,
                                              GTK_BUTTONS_OK,
                                              message.c_str ());
  gtk_dialog_run (GTK_DIALOG (dialog));
  gtk_widget_destroy (dialog);
}


void MainWindow::on_undo1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->menu_undo ();
}


void
MainWindow::menu_undo ()
{
  undo.undo (textbuffer_edit);
}


void MainWindow::on_redo1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->menu_redo ();
}


void
MainWindow::menu_redo ()
{
  undo.redo (textbuffer_edit);
}


void MainWindow::on_edit1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->menu_edit ();
}


void
MainWindow::menu_edit ()
{
  // Set the sensitivity of the items under the edit menu.
  gtk_widget_set_sensitive (undo1, undo.can_undo ());
  gtk_widget_set_sensitive (redo1, undo.can_redo ());
  // At present there does not seem to be a function that indicates
  // whether and which clipboard operations are possible. Therefore
  // the Copy, Paste and Cut menus are always active.
}


void MainWindow::on_find1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->menu_find ();
}


void
MainWindow::menu_find ()
{
  // Before finding, save the current file.
  editor_save ();
  // Start find dialog.
  bool is_indexing = false;
  {
    FindDialog finddialog (configuration);
    if (finddialog.run () == GTK_RESPONSE_OK)
      {
        is_indexing = finddialog.is_indexing;
        references_searchwords.clear ();
        references_searchwords.push_back (configuration->find_searchword);
        references_casesensitives.clear ();
        references_casesensitives.push_back (configuration->case_sensitive_search);
        sort_references (finddialog.results);
        References references (liststore, treeview1, treecolumn);
        references.set_references (finddialog.results, references_searchwords,
                                   references_casesensitives);
        references.fill_store ();
      }
  }
  /*
   * Give a message if the text is still being indexed.
   * If being indexed, the message also comes when there are none results,
   * because it might have been because of the inaccuracy that there are no
   * results, whereas when things were accurate, there could have been results.
   */
  if (is_indexing)
    {
      GtkWidget *dialog;
      dialog = gtk_message_dialog_new (GTK_WINDOW (mainwindow),
                                       GTK_DIALOG_DESTROY_WITH_PARENT,
                                       GTK_MESSAGE_WARNING, GTK_BUTTONS_OK,
                                       "The text is still being indexed.\nTherefore the results may or may not be up to date.");
      gtk_dialog_run (GTK_DIALOG (dialog));
      gtk_widget_destroy (dialog);
    }
}


void
MainWindow::on_find_and_replace1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->menu_replace ();
}


void
MainWindow::menu_replace ()
{
  // Before finding, save the current file.
  editor_save ();
  // Start find/replace dialog.
  vector < string > results;
  bool is_indexing = false;
  {
    ReplaceDialog replacedialog (configuration);
    if (replacedialog.run () == GTK_RESPONSE_OK)
      {
        is_indexing = replacedialog.is_indexing;
        references_searchwords.clear ();
        references_searchwords.push_back (configuration->find_searchword);
        references_searchwords.push_back (configuration->find_replaceword);
        references_casesensitives.clear ();
        references_casesensitives.push_back (configuration->case_sensitive_search);
        references_casesensitives.push_back (configuration->case_sensitive_search);
        sort_references (replacedialog.results);
        results.assign (replacedialog.results.begin (), replacedialog.results.end ());
        References references (liststore, treeview1, treecolumn);
        references.set_references (replacedialog.results, references_searchwords,
                                   references_casesensitives);
        references.fill_store ();
      }
    else
      {
        return;
      }
  }
  /*
   * Give a message if the text is still being indexed.
   * If being indexed, the message also comes when there are none results,
   * because it might have been because of the inaccuracy that there are no
   * results, whereas when things were accurate, there could have been results.
   */
  if (is_indexing)
    {
      GtkWidget *dialog;
      dialog = gtk_message_dialog_new (GTK_WINDOW (mainwindow),
                                       GTK_DIALOG_DESTROY_WITH_PARENT,
                                       GTK_MESSAGE_WARNING, GTK_BUTTONS_OK,
                                       "The text is still being indexed.\nTherefore the results may or may not be up to date.");
      gtk_dialog_run (GTK_DIALOG (dialog));
      gtk_widget_destroy (dialog);
    }
  // Replace text.
  if (results.size ())
    {
      string prj = configuration->project;
      ReplacingDialog replacedialog (prj, results, configuration->find_searchword,
                                     configuration->find_replaceword,
                                     configuration->case_sensitive_search);
      replacedialog.run ();
      close ();
      configuration->project = prj;
      initialize_project ();
    }
  else
    {
      GtkWidget *dialog = gtk_message_dialog_new (GTK_WINDOW (mainwindow),
                                                  GTK_DIALOG_DESTROY_WITH_PARENT,
                                                  GTK_MESSAGE_INFO,
                                                  GTK_BUTTONS_OK,
                                                  "There was nothing to replace");
      gtk_dialog_run (GTK_DIALOG (dialog));
      gtk_widget_destroy (dialog);
    }
}


void MainWindow::on_findspecial1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->menu_findspecial ();
}


void
MainWindow::menu_findspecial ()
{
  // Before finding, save the current file.
  editor_save ();
  // Start dialog.
  bool is_indexing = false;
  {
    SearchSpecialDialog dialog (configuration, liststore);
    if (dialog.run () == GTK_RESPONSE_OK)
      {
        is_indexing = dialog.is_indexing;
        references_searchwords.push_back (configuration->find_searchword);
        references_casesensitives.push_back (configuration->case_sensitive_search);
        sort_references (dialog.results);
        References references (liststore, treeview1, treecolumn);
        references.set_references (dialog.results, references_searchwords,
                                   references_casesensitives);
        references.fill_store ();
      }
  }
  /*
   * Give a message if the text is still being indexed.
   * If being indexed, the message also comes when there are none results,
   * because it might have been because of the inaccuracy that there are no
   * results, whereas when things were accurate, there could have been results.
   */
  if (is_indexing)
    {
      GtkWidget *dialog;
      dialog = gtk_message_dialog_new (GTK_WINDOW (mainwindow),
                                       GTK_DIALOG_DESTROY_WITH_PARENT,
                                       GTK_MESSAGE_WARNING, GTK_BUTTONS_OK,
                                       "The text is still being indexed.\nTherefore the results may or may not be up to date.");
      gtk_dialog_run (GTK_DIALOG (dialog));
      gtk_widget_destroy (dialog);
    }
}


void MainWindow::on_bibledit_windows_companion1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->companion_settings ();
}


void
MainWindow::companion_settings ()
{
  CompanionDialog companiondialog (configuration);
  if (companiondialog.run () == GTK_RESPONSE_OK)
    {
      // If there is a change in the local connection, deal with it.
      if (configuration->companion_local_enabled != companiondialog.local)
        {
          configuration->companion_local_enabled = companiondialog.local;
          if (configuration->companion_local_enabled)
            {
              windowscompanionlocal.Connect ("localhost", 51515);
            }
          else
            {
              windowscompanionlocal.Disconnect ();
            }
        }
      // Deal with any changes in the remote connection.
      bool b = false;
      if (configuration->companion_remote_enabled != companiondialog.remote)
        b = true;
      if (configuration->companion_remote_name.compare (companiondialog.address))
        b = true;
      if (b)
        {
          configuration->companion_remote_enabled = companiondialog.remote;
          configuration->companion_remote_name = companiondialog.address;
          if (configuration->companion_remote_enabled)
            {
              windowscompanionremote.Connect (configuration->companion_remote_name, 51515);
            }
          else
            {
              windowscompanionremote.Disconnect ();
            }
        }
    }
}


void MainWindow::on_font1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->editor_settings ();
}


void
MainWindow::editor_settings ()
{
  GtkWidget *font_selector;
  font_selector = gtk_font_selection_dialog_new ("Editor font");
  gtk_font_selection_dialog_set_font_name (GTK_FONT_SELECTION_DIALOG (font_selector),
                                           configuration->editor_font_name.c_str ());
  gtk_widget_show (font_selector);
  int result = gtk_dialog_run (GTK_DIALOG (font_selector));
  string s = gtk_font_selection_dialog_get_font_name (GTK_FONT_SELECTION_DIALOG (font_selector));
  gtk_widget_destroy (font_selector);
  if (result == GTK_RESPONSE_OK)
    {
      configuration->editor_font_name = s;
      editor_set_font ();
    }
}


void MainWindow::on_languages1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->language_settings ();
}


void
MainWindow::language_settings ()
{
  LanguagesDialog dialog (0);
  dialog.run ();
}


void MainWindow::on_stylesheets1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->stylesheet_settings ();
}


void
MainWindow::stylesheet_settings ()
{
  StylesheetsDialog dialog (0);
  dialog.run ();
}


void MainWindow::on_import1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->menu_import ();
}


void
MainWindow::menu_import ()
{
  ImportSfmDialog dialog (configuration->project);
  int result = dialog.run ();
  if (result == GTK_RESPONSE_APPLY)
    {
      string prj = configuration->project;
      close ();
      configuration->project = prj;
      initialize_project ();
    }
}


gboolean MainWindow::on_mainwindow_delete_event             (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
/*
  This solves a bug:
  When quitting the program though the menu, all went fine.
  But when quitting the program by clicking on the cross at the top of the program,
  a segmentation fault occurred.
  This segmentation occurred in the destructor of the object MainWindow.
  In the destructor properties of GTK were accessed, which obviously were 
  already in the process of being destroyed.
  This is the solution:
  When the window receives the delete_event (somebody clicks on the cross to
  close the program), the window is not closed, but instead a timeout is 
  installed that calls qtk_main_quit after a short while.
  The destructor of the object MainWindow can not access the properties of GTK
  without a segmentation fault.
  */
  g_timeout_add (10, GSourceFunc (gtk_main_quit), user_data);
  return true;
}


/*
|
|
|
|
|
Navigation
|
|
|
|
|
 */


void MainWindow::on_combo_entry_book_changed (GtkEditable * editable, gpointer user_data)
{
  ((MainWindow *) user_data)->on_combo_book_changed ();
}


void
MainWindow::on_combo_book_changed ()
{
  if (driving_book)
    return;
  string book = gtk_entry_get_text (GTK_ENTRY (combo_entry_book));
  if (book.empty ())
    return;
  go_to_reference (book, "", "", false, false);
}


void MainWindow::on_combo_entry_chapter_changed (GtkEditable * editable, gpointer user_data)
{
  ((MainWindow *) user_data)->on_combo_chapter_changed ();
}


void
MainWindow::on_combo_chapter_changed ()
{
  if (driving_chapter)
    return;
  string chapter = gtk_entry_get_text (GTK_ENTRY (combo_entry_chapter));
  if (chapter.empty ())
    return;
  go_to_reference ("", chapter, "", false, false);
}


void MainWindow::on_combo_entry_verse_changed (GtkEditable * editable, gpointer user_data)
{
  ((MainWindow *) user_data)->on_combo_verse_changed ();
}


void
MainWindow::on_combo_verse_changed ()
{
  if (driving_verse)
    return;
  string verse = gtk_entry_get_text (GTK_ENTRY (combo_entry_verse));
  if (verse.empty ())
    return;
  go_to_reference ("", "", verse, false, false);
}


void MainWindow::on_next_verse1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->goto_next_verse ();
}


void
MainWindow::goto_next_verse ()
{
  set_next_or_previous_item (GTK_COMBO (combo_verse), gtk_entry_get_text (GTK_ENTRY (combo_entry_verse)), true);
}


void MainWindow::on_previous_verse1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->goto_previous_verse ();
}


void
MainWindow::goto_previous_verse ()
{
  set_next_or_previous_item (GTK_COMBO (combo_verse), gtk_entry_get_text (GTK_ENTRY (combo_entry_verse)), false);
}


void MainWindow::on_next_chapter1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->goto_next_chapter ();
}


void
MainWindow::goto_next_chapter ()
{
  set_next_or_previous_item (GTK_COMBO (combo_chapter), gtk_entry_get_text (GTK_ENTRY (combo_entry_chapter)), true);
}


void MainWindow::on_previous_chapter1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->goto_previous_chapter ();
}


void
MainWindow::goto_previous_chapter ()
{
  set_next_or_previous_item (GTK_COMBO (combo_chapter), gtk_entry_get_text (GTK_ENTRY (combo_entry_chapter)), false);
}


void MainWindow::on_next_book1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->goto_next_book ();
}


void
MainWindow::goto_next_book ()
{
  set_next_or_previous_item (GTK_COMBO (combo_book), gtk_entry_get_text (GTK_ENTRY (combo_entry_book)), true);
}


void MainWindow::on_previous_book1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->goto_previous_book ();
}


void
MainWindow::goto_previous_book ()
{
  set_next_or_previous_item (GTK_COMBO (combo_book), gtk_entry_get_text (GTK_ENTRY (combo_entry_book)), false);
}


void MainWindow::on_reference1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->goto_reference ();
}


void
MainWindow::goto_reference ()
{
  GotoReferenceDialog dialog (navigator_current_book, navigator_current_chapter,
                              navigator_current_verse);
  if (dialog.run () == GTK_RESPONSE_OK)
    {
      if (dialog.newreference)
        go_to_reference (dialog.newbook, dialog.newchapter, dialog.newverse, true, false);
    }
}


void
MainWindow::go_to_reference (const string & gotobook, const string & gotochapter,
                             const string & gotoverse, bool jump, bool highlight)
{
  /*
   * This function makes the editor to go to a new Bible reference.
   * 
   * Going to a reference is not done immediately on request. It is instead done by a scheduler.
   * This is done because at times the program generates a lot of requests at once.
   * If the edit control is then loaded each time, and the cursor positioned,
   * this confuses Gtk when the requests come too soon one after the other.
   * So this function only stores the requested references, and a timer will carry
   * this request out later. 
   * 
   * If the previously displayed reference was edited and therefore changed,
   * it will save the file.
   * 
   * The "jump" variabele needs to be set when any arbitrary jump is made.
   * It is related to some internal logic.
   */


  // If we're busy bail out.
  if (go_to_reference_running)
    return;
  go_to_reference_running = true;

  try
  {
    // Decide where we're going to.
    string navigator_goto_book;
    if (gotobook.empty ())
      navigator_goto_book = navigator_current_book;
    else
      navigator_goto_book = gotobook;
    string navigator_goto_chapter;
    if (gotochapter.empty ())
      navigator_goto_chapter = navigator_current_chapter;
    else
      navigator_goto_chapter = gotochapter;
    string navigator_goto_verse;
    if (gotoverse.empty ())
      navigator_goto_verse = navigator_current_verse;
    else
      navigator_goto_verse = gotoverse;

    // Find out what needs to be changed: book, chapter and/or verse.
    bool new_book = (navigator_goto_book != navigator_current_book);
    bool new_chapter = (navigator_goto_chapter != navigator_current_chapter);
    bool new_verse = (navigator_goto_verse != navigator_current_verse);

    // Index of the book in the currently opened scripture.
    int bookindex = -1;
    if (scripture)
      bookindex = scripture->get_index_of_book (navigator_goto_book);

    // See whether the requested book exists.
    if (bookindex < 0)
      {
        // It doesn't exist, so do nothing.
        new_book = false;
        new_chapter = false;
        new_verse = false;
      }
    // Deal with a new book.
    if (new_book)
      {
        // Adjust the current indicator, and store the book for later.
        navigator_goto_book = scripture->books[bookindex];
        navigator_current_book = navigator_goto_book;
        configuration->book_opened = navigator_goto_book;
        // Set the text in the combo box.
        driving_book = true;
        gtk_entry_set_text (GTK_ENTRY (combo_entry_book), navigator_goto_book.c_str ());
        driving_book = false;
        // Load the book in memory. It'll provide data on the number of chapters and verses.
        if (navigator_book)
          delete navigator_book;
        navigator_book = new Book (scripture->paths[bookindex], false, false);
        // Put the chapters in the chapter combo box.
        // Also in a shadow vector, because the chapters cannot be obtained through the widget.
        GList *combo_chapter_items = NULL;
        combo_chapter_popdown_strings.clear ();
        for (unsigned int i = 1; i <= navigator_book->number_of_chapters; i++)
          {
            combo_chapter_items =
              g_list_append (combo_chapter_items, (gpointer) convert_to_string (i).c_str ());
            combo_chapter_popdown_strings.push_back (convert_to_string (i));
          }
        driving_chapter = true;
        gtk_combo_set_popdown_strings (GTK_COMBO (combo_chapter), combo_chapter_items);
        g_list_free (combo_chapter_items);
        driving_chapter = false;
        // As we've a new book, go to chapter one verse one, except when "jumping".
        if (!jump)
          {
            navigator_goto_chapter = "1";
            navigator_goto_verse = "1";
          }
        new_chapter = true;
      }
    // Bring the new chapter within the limits
    if (new_chapter)
      {
        unsigned int int_goto_chapter = convert_to_int (navigator_goto_chapter);
        if (int_goto_chapter < 1)
          navigator_goto_chapter = "1";
        if (int_goto_chapter > navigator_book->number_of_chapters)
          navigator_goto_chapter = convert_to_string (navigator_book->number_of_chapters);
      }
    // Deal with a new chapter
    if (new_chapter)
      {
        // Set the text in the combo box.
        driving_chapter = true;
        gtk_entry_set_text (GTK_ENTRY (combo_entry_chapter), navigator_goto_chapter.c_str ());
        driving_chapter = false;
        // Store the current chapter.
        navigator_current_chapter = navigator_goto_chapter;
        configuration->chapter_opened = navigator_current_chapter;
        // Put the number of verses in combo_verse
        // and in the shadow vector.
        GList *combo_verse_items = NULL;
        combo_verse_popdown_strings.clear ();
        unsigned int chapter = convert_to_int (navigator_goto_chapter);
        if (chapter > 0)
          {
            unsigned int number_of_verses = navigator_book->number_of_verses (chapter);
            for (unsigned int i = 1; i <= number_of_verses; i++)
              {
                combo_verse_items =
                  g_list_append (combo_verse_items, (gpointer) convert_to_string (i).c_str ());
                combo_verse_popdown_strings.push_back (convert_to_string (i));
              }
            driving_verse = true;
            gtk_combo_set_popdown_strings (GTK_COMBO (combo_verse), combo_verse_items);
            g_list_free (combo_verse_items);
            driving_verse = false;
          }
        // Move to beginning of the chapter, except when "jumping"
        if (!jump)
          navigator_goto_verse = "1";
        new_verse = true;
      }
    // Bring the new verse within its limits.
    if (new_verse)
      {
        int int_goto_verse = convert_to_int (navigator_goto_verse);
        int limit = navigator_book->number_of_verses (convert_to_int (navigator_goto_chapter));
        if (int_goto_verse < 1)
          navigator_goto_verse = "1";
        if (int_goto_verse > limit)
          navigator_goto_verse = convert_to_string (limit);
      }
    // Deal with a new verse.
    if (new_verse)
      {
        // Set the text in the combo box.
        driving_verse = true;
        gtk_entry_set_text (GTK_ENTRY (combo_entry_verse), navigator_goto_verse.c_str ());
        driving_verse = false;
        // Update the current verse, and store it for later also.
        navigator_current_verse = navigator_goto_verse;
        configuration->verse_opened = navigator_current_verse;
      }
    navigator_highlight = highlight;
    // If we've changed a reference: reset the timer,
    // allowing the navigation timeout to update the editor a bit later.
    if (new_book || new_chapter || new_verse)
      navigation_change_timer = 0;

  }
  catch (exception)
  {
  }

  // Ready running.
  go_to_reference_running = false;
}


bool MainWindow::mainwindow_on_navigation_timeout (gpointer data)
{
  return ((MainWindow *) data)->on_navigation_timeout ();
}


bool MainWindow::on_navigation_timeout ()
{
  /*
   * This scheduler carries out requested changes of the reference.
   * It also does related tasks.
   */

  // If F7, F8 or F9 is continually held down and it repeats,
  // we've code here to deal with that so that the interface responds
  // properly. In cases like this the change of the reference
  // is delayed until the key is released.
  navigation_change_timer++;
  if (navigation_change_timer != 4)
    return true;

  // Find out what needs to be changed: book, chapter and/or verse.
  string
    goto_book = gtk_entry_get_text (GTK_ENTRY (combo_entry_book));
  bool
    new_book = (goto_book != editor_current_book);
  editor_current_book = goto_book;
  string
    goto_chapter = gtk_entry_get_text (GTK_ENTRY (combo_entry_chapter));
  bool
    new_chapter = (goto_chapter != editor_current_chapter);
  editor_current_chapter = goto_chapter;
  string
    goto_verse = gtk_entry_get_text (GTK_ENTRY (combo_entry_verse));
  bool
    new_verse = (goto_verse != editor_current_verse);
  editor_current_verse = goto_verse;

  // See whether the current file should be saved.
  if (new_book || new_chapter)
    if (gtk_text_buffer_get_modified (textbuffer_edit))
      editor_save ();

  // Deal with a new book.
  if (new_book)
    {
      int
        bookindex = scripture->get_index_of_book (goto_book);
      // Load book, do not split the lines.
      if (book)
        delete
          book;
      book = new Book (scripture->paths[bookindex], false, true);
      // With a new book, also load a new chapter.
      new_chapter = true;
    }
  // Deal with a new chapter.
  if (new_chapter)
    {
      // Load chapter, if need be.
      editor_load_chapter ();
      // When loading a new chapter, we've also a new verse.
      new_verse = true;
    }

  if (new_book || new_chapter || new_verse)
    {
      // Position the cursor properly.
      // The positioning will be done whenever Gtk is idle.
      // This is because sometimes Gtk is slow in loading a new chapter.
      // So if the cursor positioning is done straight after loading,
      // it might not work.
      goto_cursor_position = goto_verse;
      g_idle_add (GSourceFunc (mainwindow_editor_position_cursor_at_verse), gpointer(this));
      // Send the reference to the windows companions.
      unsigned int
        index = english_id_to_index (goto_book);
      string
        value =
        "gotoreference " + index_to_paratext_id (index) + " " + goto_chapter + ":" + goto_verse;
      windowscompanionlocal.Sendln (value);
      windowscompanionremote.Sendln (value);
    }
  return true;
}


void MainWindow::on_textview_edit_move_cursor (GtkTextView * textview, GtkMovementStep step, gint count,
                              gboolean extend_selection, gpointer user_data)
{
  // This function is called _before_ the actual move takes place, but we
  // do need a function called _after_ the move. To solve that here we
  // call the function after a short delay, and assume that the move
  // has taken place within that delay.
  g_timeout_add (1, GSourceFunc (mainwindow_on_cursor_moved_delayed), user_data);
}


bool MainWindow::mainwindow_on_cursor_moved_delayed (gpointer data)
{
  ((MainWindow *) data)->on_cursor_moved ();
  return false;
}


void
MainWindow::on_cursor_moved ()
{
  /*
   * When the cursor has moved, the navigation system needs to be updated
   * so that it shows the right reference. If the user was, for example
   * on MAT 1:10, and the cursor moves, the move might have brought him or her
   * to another reference, though this is not necessarily so. Therefore, as we 
   * don't know where the user is now after the cursor moved, we need to find
   * it out. The book is known, the chapter is known, because both stay the same.
   * The only thing we don't know is the verse. 
   */

  // Somebody thought that in gtkmm 2.4 the move_cursor signal was useless,
  // so removed it. As we need it here we we trying to go back to the 
  // underlying GTK2 object, and connect the signal from there.
  // on_cursor_moved_tag = g_signal_connect(textview.gobj(), "move_cursor", on_move_cursor, 0);
  // on_cursor_moved_tag = g_signal_connect(textview.gobj(), "move_cursor", *this.on_cursor_moved(), 0);
  // But this does not seem to work.
  // I feel a bit alienated from the gtkmm library because of this,
  // and consider moving to the GTK2 library.

/*
  gtkmm 2.4 removed signal_cursor_moved, so bibledit is now polling the
  line the cursor is on, and takes action when it has changed.
  */

/*
 * The above is partly true now, because we've moved to gtk+ and threw out gtkmm.
 * The polling has gone too, and the function is now called only when the cursor moved.
 */

  static int previous_cursor_line = 0;
  /* Initializing the value to 0 instead of -1 solves a Segmentation fault 
   * exception when no document was loaded.
   */
  GtkTextIter iterator;
  gtk_text_buffer_get_iter_at_mark (textbuffer_edit, &iterator,
                                    gtk_text_buffer_get_insert (textbuffer_edit));
  int line_of_cursor = gtk_text_iter_get_line (&iterator);
  if (line_of_cursor != previous_cursor_line)
    {
      previous_cursor_line = line_of_cursor;
      highlight_verse_at_cursor ();
      go_to_reference ("", "", cursor_get_verse_number (), false, false);
    }
}


void MainWindow::on_textview_edit_grab_focus (GtkWidget * widget, gpointer user_data)
{
  ((MainWindow *) user_data)->on_grab_focus ();
}


void
MainWindow::on_grab_focus ()
{
  // The purpose of this is to adjust the reference to the new position,
  // wherever the user clicks in the text.
  if (focus_programmatically_being_grabbed)
    return;
  g_idle_add (GSourceFunc (mainwindow_on_grab_focus_delayed_handler), gpointer(this));
}


string MainWindow::cursor_get_verse_number ()
{
  /*
   * This gets the verse number the cursor is on.
   * We will find the verse out by doing the following things:
   * 1. Get the line the cursor is on. If it has a verse marker, we then can get
   * the verse number from there.
   * 2. If we did not find a verse number, get the previous line, and see what
   * information is there.
   * 3. Keep doing step 2 until we either have the information we want, or we have
   * arrived at the beginning of the file.
   * 4. If we are at the start of the file, assume verse one.
   */
  // Get the line the cursor is on.
  GtkTextIter
    iterator;
  gtk_text_buffer_get_iter_at_mark (textbuffer_edit, &iterator,
                                    gtk_text_buffer_get_insert (textbuffer_edit));
  int
    line_of_cursor = gtk_text_iter_get_line (&iterator);
  // If no verse information is going to be found, assume verse one.
  string
    verse = "1";
  // Go through all lines, from the cursor downwards, and check whether verse information is there.
  GtkTextIter
    endofline;
  Usfm
  usfm (0);
  for (int i = line_of_cursor; i >= 0; i--)
    {
      gtk_text_buffer_get_iter_at_line (textbuffer_edit, &iterator, i);
      gtk_text_buffer_get_iter_at_line (textbuffer_edit, &endofline, i + 1);
      gtk_text_iter_backward_char (&endofline);
      ustring
        s = gtk_text_buffer_get_text (textbuffer_edit, &iterator, &endofline, false);
      ustring
        marker = usfm_extract_marker (s);
      if (usfm.is_verse (marker))
        {
          verse = number_in_string (s);
          break;
        }
    }
  return verse;
}


bool MainWindow::mainwindow_editor_position_cursor_at_verse (gpointer data)
{
  return ((MainWindow *) data)->editor_position_cursor_at_verse ();
}


bool MainWindow::editor_position_cursor_at_verse ()
{
  /*
   * This function moves the cursor of the editor to the verse stored before.
   */

  // We will not move the cursor or scroll to it when the cursor is already on the right verse.
  bool
    b = (goto_cursor_position.compare (cursor_get_verse_number ()));
  // When the cursor is at the end of the buffer, it usually means that the buffer has just been filled.
  // So the cursor does need to be repositioned in this case.
  GtkTextIter
    cursor_iterator;
  gtk_text_buffer_get_iter_at_mark (textbuffer_edit, &cursor_iterator,
                                    gtk_text_buffer_get_insert (textbuffer_edit));
  GtkTextIter
    enditer;
  gtk_text_buffer_get_end_iter (textbuffer_edit, &enditer);
  if (gtk_text_iter_compare (&cursor_iterator, &enditer) == 0)
    b = true;
  if (b)
    {
      // Grab focus here to get the scrolling done properly, and the user can type in it.
      focus_programmatically_being_grabbed = true;
      gtk_widget_grab_focus (textview_edit);
      focus_programmatically_being_grabbed = false;
      // Go through all the lines and find out about the verse.
      Usfm
      usfm (0);
      for (int i = 0; i < gtk_text_buffer_get_line_count (textbuffer_edit); i++)
        {
          GtkTextIter
            begin;
          gtk_text_buffer_get_iter_at_line (textbuffer_edit, &begin, i);
          GtkTextIter
            end;
          gtk_text_buffer_get_iter_at_line (textbuffer_edit, &end, i + 1);
          ustring
            line = gtk_text_buffer_get_text (textbuffer_edit, &begin, &end, false);
          ustring
            marker = usfm_extract_marker (line);
          if (usfm.is_verse (marker))
            {
              ustring
                verse = number_in_string (line);
              if (verse == goto_cursor_position)
                {
                  // Move the cursor to it.
                  // Move it to the beginning of the text, if there is any.
                  int
                    offset = 3 + marker.length () + verse.length ();
                  ustring
                    s = gtk_text_buffer_get_text (textbuffer_edit, &begin, &end, false);
                  int
                    linelength = s.length ();
                  linelength--;
                  if (offset >= linelength)
                    offset = linelength;
                  gtk_text_buffer_get_iter_at_line_offset (textbuffer_edit, &begin, i, offset);
                  gtk_text_buffer_place_cursor (textbuffer_edit, &begin);
                  // Scroll also to it. It will scroll to the beginning of the text after the verse marker.
                  // Alignment is needed to put the line being edited near the top of the window.
                  gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW (textview_edit), &begin, 0.1, false,
                                                0, 0);
                  // Highlight the verse we moved to (remove any previous tags)
                  highlight_verse_at_cursor ();
                  break;
                }
            }
        }
      // Highlight searchwords if need be.
      highlight_searchwords (convert_to_int (goto_cursor_position));
    }
  return false;
}


/*
|
|
|
|
|
Tracker
|
|
|
|
|
 */


void MainWindow::on_button_forward_clicked (GtkButton * button, gpointer user_data)
{
  ((MainWindow *) user_data)->goto_next_reference ();
}


void
MainWindow::goto_next_reference ()
{
  set_next_or_previous_reference (true);
}


void MainWindow::on_button_back_clicked (GtkButton * button, gpointer user_data)
{
  ((MainWindow *) user_data)->goto_previous_reference ();
}


void
MainWindow::goto_previous_reference ()
{
  set_next_or_previous_reference (false);
}


void
MainWindow::set_next_or_previous_item (GtkCombo * combo, const string & current_text, bool next)
{
  // This function takes the next (or previous) value in the dropdown list of the combo box
  // and positions the editor there.
  vector < string > strings;
  if (GTK_WIDGET (combo) == combo_book)
    for (unsigned int i = 0; i < combo_book_popdown_strings.size (); i++)
      strings.push_back (combo_book_popdown_strings[i]);
  if (GTK_WIDGET (combo) == combo_chapter)
    for (unsigned int i = 0; i < combo_chapter_popdown_strings.size (); i++)
      strings.push_back (combo_chapter_popdown_strings[i]);
  if (GTK_WIDGET (combo) == combo_verse)
    for (unsigned int i = 0; i < combo_verse_popdown_strings.size (); i++)
      strings.push_back (combo_verse_popdown_strings[i]);
  if (strings.size () == 0)
    return;
  unsigned int index = 0;
  for (unsigned int i = 0; i < strings.size (); i++)
    {
      if (current_text == strings[i])
        index = i;
    }
  if (next)
    {
      if (index == (strings.size () - 1))
        return;
      index++;
    }
  else
    {
      if (index == 0)
        return;
      index--;
    }
  string combo_text;
  combo_text = strings[index];
  // Go to the new reference.
  if (GTK_WIDGET (combo) == combo_book)
    go_to_reference (combo_text, "", "", false, false);
  if (GTK_WIDGET (combo) == combo_chapter)
    go_to_reference ("", combo_text, "", false, false);
  if (GTK_WIDGET (combo) == combo_verse)
    go_to_reference ("", "", combo_text, false, false);
}


void
MainWindow::set_sensitivity_of_navigator_buttons ()
{
  // Enables or disables each navigator button, depending on whether references for it are available.
  gtk_widget_set_sensitive (button_back, track.previous_reference_available ());
  gtk_widget_set_sensitive (button_forward, track.next_reference_available ());
}


bool MainWindow::mainwindow_on_tracker_timeout (gpointer data)
{
  return ((MainWindow *) data)->on_tracker_timeout ();
}


bool MainWindow::on_tracker_timeout ()
{
  // Drive the Bible reference tracker.
  track.tick (gtk_entry_get_text (GTK_ENTRY (combo_entry_book)),
              gtk_entry_get_text (GTK_ENTRY (combo_entry_chapter)),
              gtk_entry_get_text (GTK_ENTRY (combo_entry_verse)));
  set_sensitivity_of_navigator_buttons ();
  return true;
}


void
MainWindow::set_next_or_previous_reference (bool next)
{
  string book;
  string chapter;
  string verse;
  if (next)
    track.get_next_reference (book, chapter, verse);
  else
    track.get_previous_reference (book, chapter, verse);
  go_to_reference (book, chapter, verse, true, false);
  set_sensitivity_of_navigator_buttons ();
}


bool MainWindow::mainwindow_on_grab_focus_delayed_handler (gpointer data)
{
  return ((MainWindow *) data)->on_grab_focus_delayed_handler ();
}


bool MainWindow::on_grab_focus_delayed_handler ()
{
  /*
   * If the user clicks in the editor window, and straight after that
   * the verse where the cursor is, is asked, we get a reply 
   * indicating the verse where the cursor was prior to clicking.
   * This delayed handler solves that.
   */
  go_to_reference ("", "", cursor_get_verse_number (), false, false);
  highlight_verse_at_cursor ();
  return false;
}


void
MainWindow::highlight_verse_at_cursor ()
{
  // Remove previous highlight, if any.
  GtkTextIter startiter;
  GtkTextIter enditer;
  gtk_text_buffer_get_start_iter (textbuffer_edit, &startiter);
  gtk_text_buffer_get_end_iter (textbuffer_edit, &enditer);
  gtk_text_buffer_remove_tag (textbuffer_edit, editor_line_tag, &startiter, &enditer);
  // Hightlight the line number the cursor is on.
  gtk_text_buffer_get_iter_at_mark (textbuffer_edit, &startiter,
                                    gtk_text_buffer_get_insert (textbuffer_edit));
  int line_of_cursor = gtk_text_iter_get_line (&startiter);
  gtk_text_buffer_get_iter_at_line (textbuffer_edit, &startiter, line_of_cursor);
  gtk_text_buffer_get_iter_at_line (textbuffer_edit, &enditer, ++line_of_cursor);
  gtk_text_buffer_apply_tag (textbuffer_edit, editor_line_tag, &startiter, &enditer);
}


/*
|
|
|
|
|
Editing
|
|
|
|
|
 */


void
MainWindow::editor_load_chapter ()
{
  /*
   * Loads the chapter in the editor.
   */
  // Set the new text.
  vector < ustring > lines;
  book->get_chapter (convert_to_int (gtk_entry_get_text (GTK_ENTRY (combo_entry_chapter))), lines);
  ustring text;
  for (unsigned int i = 0; i < lines.size (); i++)
    {
      text.append (lines[i] + "\n");
    }
  gtk_text_buffer_set_text (textbuffer_edit, text.c_str (), -1);
  // Set the editor ummodified.
  gtk_text_buffer_set_modified (textbuffer_edit, false);
  // Clear the undo / redo stack.
  undo.clear ();
}


void
MainWindow::editor_save ()
{
  /*
   * This will save the book if it has been changed in the editor.
   */
  // See whether it was changed.
  if (gtk_text_buffer_get_modified (textbuffer_edit))
    {
      // Collect all lines of the editor
      vector < ustring > lines;
      int number_of_lines = gtk_text_buffer_get_line_count (textbuffer_edit);
      GtkTextIter iterator;
      GtkTextIter endofline;
      for (int i = 0; i < number_of_lines; i++)
        {
          gtk_text_buffer_get_iter_at_line (textbuffer_edit, &iterator, i);
          gtk_text_buffer_get_iter_at_line (textbuffer_edit, &endofline, i + 1);
          gtk_text_iter_backward_char (&endofline);
          string s = gtk_text_buffer_get_text (textbuffer_edit,
                                               &iterator,
                                               &endofline,
                                               false);
          s = trim (s);
          if (!s.empty ())
            lines.push_back (s);
        }
      // Rejoin the lines in the chapter and save the whole book.
      book->set_chapter (lines);
      book->save ();
      // Set the buffer to unmodified.
      gtk_text_buffer_set_modified (textbuffer_edit, false);
      // Start the process that will update the word database.
      string command = "bibledit.wdb --update ";
      command.append (configuration->project);
      command.append (" '");
      command.append (index_to_biblebook (book->get_id_index ()));
      command.append ("' ");
      command.append (convert_to_string ((int) book->get_previously_presented_chapter ()));
      Shell shell (command);
      shell.set_asynchronous ();
      shell.run ();
    }
}


bool MainWindow::mainwindow_on_save_timeout (gpointer data)
{
  return ((MainWindow *) data)->on_save_timeout ();
}


bool MainWindow::on_save_timeout ()
{
  editor_save ();
  return true;
}


void
MainWindow::editor_set_font ()
{
  PangoFontDescription *font_desc = NULL;
  font_desc = pango_font_description_from_string (configuration->editor_font_name.c_str ());
  gtk_widget_modify_font (textview_edit, font_desc);
  pango_font_description_free (font_desc);
}


/*
|
|
|
|
|
Clipboard
|
|
|
|
|
 */


void MainWindow::on_cut1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->on_cut ();
}


void
MainWindow::on_cut ()
{
  GtkClipboard *clipboard;
  clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
  if (GTK_WIDGET_HAS_FOCUS (textview_edit))
    gtk_text_buffer_cut_clipboard (textbuffer_edit, clipboard, true);
}


void MainWindow::on_copy1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->on_copy ();
}


void
MainWindow::on_copy ()
{
  GtkClipboard *clipboard;
  clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
  if (GTK_WIDGET_HAS_FOCUS (textview_edit))
    gtk_text_buffer_copy_clipboard (textbuffer_edit, clipboard);
}


void MainWindow::on_paste1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->on_paste ();
}


void
MainWindow::on_paste ()
{
  GtkClipboard *clipboard;
  clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
  if (GTK_WIDGET_HAS_FOCUS (textview_edit))
    gtk_text_buffer_paste_clipboard (textbuffer_edit, clipboard, NULL, true);
}


/*
|
|
|
|
|
Backups
|
|
|
|
|
 */


/*
|
|
|
|
|
List store and reference handling
|
|
|
|
|
 */


gboolean MainWindow::on_treeview1_key_press_event (GtkWidget * widget, GdkEventKey * event, gpointer user_data)
{
  /* Pressing Return on the keyboard, or Enter on the numerical keypad
   * make us go to the reference.
   */
  if (event->keyval == GDK_Return || event->keyval == GDK_KP_Enter)
    ((MainWindow *) user_data)->on_list_goto ();
  // Pressing Delete takes the reference(s) out that have been selected.
  if (event->keyval == GDK_Delete || event->keyval == GDK_KP_Delete)
    ((MainWindow *) user_data)->on_delete_references ();
  return FALSE;
}


gboolean MainWindow::on_treeview1_button_press_event (GtkWidget * widget, GdkEventButton * event, gpointer user_data)
{
  // Double-clicking a references makes us go to the reference.
  if (event->type == GDK_2BUTTON_PRESS)
    ((MainWindow *) user_data)->on_list_goto ();
  return FALSE;
}


void
MainWindow::on_list_goto ()
{
  /*
   * Handler for when the user pressed Enter in the list and wants to go to a reference.
   */
  try
  {
    // Get the string the user selected.
    GtkTreeSelection *selection;
    selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview1));
    gtk_tree_selection_selected_foreach (selection, mainwindow_selection_foreach_function, gpointer(this));
    // Decode into book, chapter and verse.
    string book;
    string chapter;
    string verse;
    decode_reference (selected_reference, book, chapter, verse);
    // Jump to the reference.
    go_to_reference (book, chapter, verse, true, true);
  }
  catch (exception & ex)
  {
    cerr << ex.what () << endl;
  }
}


void MainWindow::mainwindow_selection_foreach_function (GtkTreeModel * model, GtkTreePath * path, GtkTreeIter * iter, gpointer data)
{
  gchar *book;
  gtk_tree_model_get (model, iter, 0, &book, -1);
  ((MainWindow *) data)->selected_reference = book;
  g_free (book);
}


void MainWindow::on_open_references1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->on_open_references ();
}


void
MainWindow::on_open_references ()
{
  GtkWidget *file_selector;
  file_selector = gtk_file_selection_new ("Please choose a file");
  gtk_file_selection_hide_fileop_buttons (GTK_FILE_SELECTION (file_selector));
  gtk_file_selection_set_filename (GTK_FILE_SELECTION (file_selector),
                                   configuration->references_file.c_str ());
  gtk_widget_show (file_selector);
  int result = gtk_dialog_run (GTK_DIALOG (file_selector));
  string s = gtk_file_selection_get_filename (GTK_FILE_SELECTION (file_selector));
  gtk_widget_destroy (file_selector);
  if (result == GTK_RESPONSE_OK)
    {
      configuration->references_file = s;
      References references (liststore, treeview1, treecolumn);
      references.load (configuration->references_file);
      references.get_references (references_searchwords, references_casesensitives);
      references.fill_store ();
    }
}


void MainWindow::on_references_save_as_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->on_save_references ();
}


void
MainWindow::on_save_references ()
{
  try
  {
    GtkWidget *file_selector;
    file_selector = gtk_file_selection_new ("Please choose a file");
    gtk_file_selection_hide_fileop_buttons (GTK_FILE_SELECTION (file_selector));
    gtk_file_selection_set_filename (GTK_FILE_SELECTION (file_selector),
                                     configuration->references_file.c_str ());
    gtk_widget_show (file_selector);
    int result = gtk_dialog_run (GTK_DIALOG (file_selector));
    string s = gtk_file_selection_get_filename (GTK_FILE_SELECTION (file_selector));
    gtk_widget_destroy (file_selector);
    if (result == GTK_RESPONSE_OK)
      {
        configuration->references_file = s;
        References references (liststore, treeview1, treecolumn);
        // Hack: set references with a dummy, then load the real ones from the editor.
        vector < string > dummy;
        references.set_references (dummy, references_searchwords, references_casesensitives);
        references.get_loaded ();
        references.save (s);
      }
  }
  catch (exception & ex)
  {
    cerr << "Saving references: " << ex.what () << endl;
  }
}


void MainWindow::on_close_references_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->on_clear_references ();
}


void
MainWindow::on_clear_references ()
{
  References references (liststore, treeview1, treecolumn);
  references.fill_store ();
}


void MainWindow::on_delete_references_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->on_delete_references ();
}


void
MainWindow::on_delete_references ()
{
  // Delete each selected row.
  GtkTreeSelection *selection;
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview1));
  vector < GtkTreeIter > iters;
  gtk_tree_selection_selected_foreach (selection, MainWindow::on_delete_reference,
                                       gpointer (&iters));
  for (unsigned int i = 0; i < iters.size (); i++)
    {
      GtkTreeIter iter = iters[i];
      gtk_list_store_remove (liststore, &iter);
    }
  // After deleting reference(s) save any remaining references to disk.
  References references (liststore, treeview1, treecolumn);
  references.get_loaded ();
  references.save ("");
  // Update the header with the number of references.
  references.set_header ();
}


void
MainWindow::on_delete_reference (GtkTreeModel * model, GtkTreePath * path, GtkTreeIter * iter,
                                 gpointer data)
{
  ((vector < GtkTreeIter > *)data)->push_back (*iter);
}


void MainWindow::on_next_reference1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->on_next_reference ();
}


void
MainWindow::on_next_reference ()
{
  /* 
   * This goes to the next reference, if there is any.
   * If no item has been selected it chooses the first, if it's there.
   */
  // Select next reference in the list.
  References references (liststore, treeview1, treecolumn);
  references.goto_next ();
  // Actually open the reference in the editor.
  on_list_goto ();
}


void MainWindow::on_previous_reference1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  ((MainWindow *) user_data)->on_previous_reference ();
}


void
MainWindow::on_previous_reference ()
{
  /* 
   * This goes to the previous reference, if there is any.
   * If no item has been selected it chooses the first, if it's there.
   */
  // Select previous reference in the list.
  References references (liststore, treeview1, treecolumn);
  references.goto_previous ();
  // Actually open the reference in the editor.
  on_list_goto ();
}


void
MainWindow::highlight_searchwords (int versenumber)
{
/*
	This highlights all the words in the verse that agree to the word
	we search for.
*/
  if (navigator_highlight)
    {
      // Remove previous highlight, if any.
      GtkTextIter startiter;
      GtkTextIter enditer;
      gtk_text_buffer_get_start_iter (textbuffer_edit, &startiter);
      gtk_text_buffer_get_end_iter (textbuffer_edit, &enditer);
      gtk_text_buffer_remove_tag (textbuffer_edit, editor_reference_tag, &startiter, &enditer);
      // The word(s) to search for.
      for (unsigned int i = 0; i < references_searchwords.size (); i++)
        {
          ustring searchword = references_searchwords[i];
          bool casesensitive = references_casesensitives[i];
          if (!casesensitive)
            searchword = searchword.casefold ();
          // Highlight words in the line the cursor is on.
          bool cursor_set = false;
          GtkTextIter iterator;
          gtk_text_buffer_get_iter_at_mark (textbuffer_edit, &iterator,
                                            gtk_text_buffer_get_insert (textbuffer_edit));
          int line_of_cursor = gtk_text_iter_get_line (&iterator);
          bool is_verse;
          highlight_searchwords_internal (line_of_cursor, versenumber, searchword, cursor_set,
                                          is_verse, casesensitive);
          // See whether we have lines before this verse that is not a verse number,
          // and highlight words there too.
          int lower_bound = line_of_cursor - 5;
          lower_bound = CLAMP (lower_bound, 0, line_of_cursor);
          int higher_bound = line_of_cursor - 1;
          higher_bound = CLAMP (higher_bound, 0, line_of_cursor);
          is_verse = false;
          for (int i = higher_bound; i >= lower_bound; i--)
            if (!is_verse)
              highlight_searchwords_internal (i, versenumber, searchword, cursor_set, is_verse,
                                              casesensitive);
          // Do the same for some lines after this verse.
          int hi_line = gtk_text_buffer_get_line_count (textbuffer_edit) - 1;
          lower_bound = line_of_cursor + 1;
          lower_bound = CLAMP (lower_bound, line_of_cursor, hi_line);
          higher_bound = line_of_cursor + 5;
          higher_bound = CLAMP (higher_bound, line_of_cursor, hi_line);
          is_verse = false;
          for (int i = lower_bound; i <= higher_bound; i++)
            if (!is_verse)
              highlight_searchwords_internal (i, versenumber, searchword, cursor_set, is_verse,
                                              casesensitive);
        }
    }
}


void
MainWindow::highlight_searchwords_internal (unsigned int linenumber, unsigned int versenumber,
                                            const string & searchword, bool & cursor_set,
                                            bool & is_verse, bool casesensitive)
{
  // Extract the line.
  GtkTextIter begin;
  gtk_text_buffer_get_iter_at_line (textbuffer_edit, &begin, linenumber);
  GtkTextIter end;
  gtk_text_buffer_get_iter_at_line (textbuffer_edit, &end, linenumber + 1);
  ustring line = gtk_text_buffer_get_text (textbuffer_edit, &begin, &end, false);
  if (!casesensitive)
    line = line.casefold ();
  // Get the verse number.
  string this_verse;
  Usfm usfm (0);
  ustring marker (line);
  marker = usfm_extract_marker (marker);
  is_verse = usfm.is_verse (marker);
  if (is_verse)
    this_verse = number_in_string (line);
  // If there is any verse number and it is different from the one we have, return.
  if (is_verse)
    if (convert_to_int (this_verse) != versenumber)
      return;
  // Find all places in this line that have the search word.
  unsigned int offposition = line.find (searchword);
  while (offposition != string::npos)
    {
      // Highlight the word.
      gtk_text_buffer_get_iter_at_line_offset (textbuffer_edit, &begin, linenumber, offposition);
      gtk_text_buffer_get_iter_at_line_offset (textbuffer_edit, &end, linenumber,
                                               offposition + searchword.length ());
      gtk_text_buffer_apply_tag (textbuffer_edit, editor_reference_tag, &begin, &end);
      if (!cursor_set)
        {
          gtk_text_buffer_place_cursor (textbuffer_edit, &begin);
          cursor_set = true;
        }
      offposition = line.find (searchword, offposition + searchword.length ());
    }
}


/*
|
|
|
|
|
Undo / Redo
|
|
|
|
|
 */


void MainWindow::on_textbuffer_edit_insert_text (GtkTextBuffer * textbuffer, GtkTextIter * arg1, gchar * arg2,
                                gint arg3, gpointer user_data)
{
  // As the documentation of this callback is not so good (gtk+ 2.6 was checked),
  // here follows some extra info.
  // textbuffer is the buffer where the text was inserted.
  // arg1 is the iterator pointing to the location where the insert was done.
  // arg2 is the string that was inserted.
  // arg3 is the length of the string that was inserted.
  ustring text (arg2);
  ((MainWindow *) user_data)->on_insert (textbuffer, arg1, text, text.length ());
}


void MainWindow::on_textbuffer_edit_delete_text (GtkTextBuffer * textbuffer, GtkTextIter * arg1, GtkTextIter * arg2,
                                gpointer user_data)
{
  // As the documentation of this callback is not so good (gtk+ 2.6 was checked),
  // here follows some extra info.
  // textbuffer is the buffer where text was deleted from.
  // arg1 is the iterator pointing to the beginning of the deleted text.
  // arg2 is the iterator pointing to the end of the deleted text.
  ((MainWindow *) user_data)->on_erase (textbuffer, arg1, arg2);
}


void
MainWindow::on_insert (GtkTextBuffer * textbuffer, GtkTextIter * pos, const ustring & text,
                       int bytes)
{
  undo.on_insert (pos, text, bytes);
}


void
MainWindow::on_erase (GtkTextBuffer * textbuffer, GtkTextIter * range_begin,
                      GtkTextIter * range_end)
{
  undo.on_erase (range_begin, range_end);
}


/*
|
|
|
|
|
Bibledit Windows Companion
|
|
|
|
|
 */


bool MainWindow::mainwindow_on_windows_companion_timeout (gpointer data)
{
  return ((MainWindow *) data)->on_windows_companion_timeout ();
}


bool MainWindow::on_windows_companion_timeout ()
{
  if (configuration->companion_local_enabled)
    windowscompanionlocal.Tick ();
  if (configuration->companion_remote_enabled)
    windowscompanionremote.Tick ();
  return true;
}


/*
|
|
|
|
|
Test development
|
|
|
|
|
 */


void
MainWindow::on_test ()
{
}
